#include "Canvas.h"

#include "allegro.h"

#include "RectMesh.h"


#include "sys_include.h"


using namespace std;


// TODO: replace the + 0.5 with something that can handle negative numbers also.

// completely around the picture. pixel center is at the coordinate 0, 0.

// pixel edge. 0, 0 is edge.

Canvas::Canvas(): mpPic(NULL), mAllegParallel(false), mDebugLevel(0)
{
}


void Canvas::SetDebugLevel(int level)
{
    mDebugLevel = level;
}


void Canvas::Coverage(RectMesh* pMesh, float x, float y)
{
    assert(pMesh);
    
    //pMesh->AddRect(-0.5f + x, -0.5f + y, Xsize(), Ysize());
    pMesh->AddRect(x, y, Xsize(), Ysize());
}


void Canvas::Pixel(float x, float y, int r, int g, int b, int a)
{
	assert(mpPic);
	putpixel(mpPic, int(x + 0.5f), int(y + 0.5f), makeacol(r, g, b, a));
}


void Canvas::SetAllegro(BITMAP *pic)
{
	if (!mAllegParallel && mpPic) destroy_bitmap(mpPic);
	mpPic = pic;
	mAllegParallel = false;
}


void Canvas::SetAllegroPointer(BITMAP *pic)
{
	if (!mAllegParallel && mpPic) destroy_bitmap(mpPic);
	mpPic = pic;
	mAllegParallel = true;
}


void Canvas::SetSize(float xSize, float ySize)
{
	assert((xSize + 0.5) >= 0 && (ySize + 0.5) >= 0);
	assert(!mAllegParallel);

	// allegro might crash if a bitmap with xSize * ySize == 0
	if (mpPic) {
		destroy_bitmap(mpPic);
		mpPic = NULL;
	}

	if (xSize >= 0.5f && ySize >= 0.5f) mpPic = create_bitmap_ex(32, int(xSize + 0.5), int(ySize + 0.5));
	//if (xSize >= 0.5f && ySize >= 0.5f) mpPic = create_bitmap(int(xSize + 0.5f), int(ySize + 0.5f));
}


void Canvas::Clear(int r, int g, int b, int a)
{
	if (mpPic) clear_to_color(mpPic, makeacol(r, g, b, a));
}


float Canvas::Xsize(void)
{
	if (mpPic) return mpPic->w;
		else return 0;
}


float Canvas::Ysize(void)
{
	if (mpPic) return mpPic->h;
		else return 0;
}


BITMAP* Canvas::mpGetAllegro(void)
{
	assert(NULL != mpPic); 
	return mpPic;
}


void Canvas::Load(const string& rFname)
{
    PALETTE pal; // this will be just dumped. no support for 256 color displays.
	if (!mAllegParallel && mpPic) { destroy_bitmap(mpPic); }
	mAllegParallel = false;


    assert(4 <= rFname.size()); // has to have proper extension.
    string ext = rFname.substr(rFname.size() - 4, 4);

	//loginf << "ext : " << ext << endl;
	if (".png" == ext || ".PNG" == ext) { // is a png
		// TODO: what are these for?
set_color_conversion(COLORCONV_KEEP_TRANS); // or COLORCONV_NONE
//	set_color_conversion(COLORCONV_KEEP_TRANS | COLORCONV_8_TO_32 | COLORCONV_15_TO_32 | COLORCONV_16_TO_32 | COLORCONV_24_TO_32);
//	set_color_conversion(COLORCONV_8_TO_32 | COLORCONV_15_TO_32 | COLORCONV_16_TO_32 | COLORCONV_24_TO_32);
        //set_color_conversion(COLORCONV_NONE); // or COLORCONV_KEEP_TRANS // hmh. keep_trans slow ;( whaat?
//         set_color_conversion(COLORCONV_TOTAL); // total-alpha-cut :)
        BlendMode(EMulti);
	} else { // is not a png file. so, assume no alpha.
        set_color_conversion(COLORCONV_TOTAL);
        BlendMode(ECopy);
	}

	
	mpPic = load_bitmap(rFname.c_str(), pal);
	if (!mpPic) { 
		logerr << "Canvas loading failed. filename: " << rFname << endl;
		exit(1);
		
		// Don't remember if this works. Better not to try.
		// memory problems..
		mpPic = create_bitmap(20, 20);
		clear_to_color(mpPic, makeacol(0, 255, 0, 255));
	}

	// logic error. allegro error. whatever. no other depths supported.
	//loginf << endl << "filename: " << rFname << endl;
	//LOG(bitmap_color_depth(mpPic));
    if (32 != bitmap_color_depth(mpPic)) {
        logerr << "wrong colordepth. internal error. file: " << rFname
               << endl << "depth: " << bitmap_color_depth(mpPic) << endl
               << "should be 32" << endl;
        exit(2);
    }

	//assert(32 == bitmap_color_depth(mpPic));
}


// pixel edge


// The mask is NOT on *this* Canvas.
void Canvas::Draw(Canvas* pDest, RectMesh* pMask, float upLx, float upLy)
{
//loginf << endl << endl;

    if (NULL == pMask)
    {
        Canvas::Draw(pDest, upLx, upLy);
        return;
    }

    if (0 >= Xsize() || 0 >= Ysize()) return;

	if (ECopy == BlendMode())
	{
        // *this bounding rectangle.
        Recta th;
        th.Pos(upLx, upLy);
        th.Size(Xsize(), Ysize());
        
        for (RectMesh::iterator i = pMask->begin(); i != pMask->end(); i++)
        {
			// Some kind of sub-pixel accuracy. hopefully!!

            Recta r = *i;

            if (!r.Clip(th)) continue; // They don't even overlap. nothing to draw.

			int screen_x = (int)floor(r.Ulx() + 0.5f);
			int screen_y = (int)floor(r.Uly() + 0.5f);

			int size_x = int(floor(r.Ulx() + r.Xsize() + 0.5f)) - screen_x;
			int size_y = int(floor(r.Uly() + r.Ysize() + 0.5f)) - screen_y;

//			int pic_x = (int)floor(screen_x - upLx);
//			int pic_y = (int)floor(screen_y - upLy);
			int pic_x = (int)floor(screen_x - upLx + 0.5f);
			int pic_y = (int)floor(screen_y - upLy + 0.5f);

            
			// Sub-pixel stuff.
            if (0 >= size_x || 0 >= size_y) continue; // nothing to draw.
			
			
            if (-1 == mDebugLevel || 1 <= mDebugLevel)
                mDebugDraw(pDest,
                           screen_x,
                           screen_y,
                           size_x, size_y);

            else

			blit(mpGetAllegro(), pDest->mpGetAllegro(),
                 pic_x,
                 pic_y,
                 screen_x, screen_y,
                 size_x, size_y);
        }
        
        return;
	}

    if (EMulti == BlendMode())
    {
        // *this bounding rectangle.
        Recta th;
        th.Pos(upLx, upLy);
        th.Size(Xsize(), Ysize());

		for (RectMesh::iterator i = pMask->begin(); i != pMask->end(); i++)
        {
            Recta r = *i;

            if (!r.Clip(th)) continue; // They don't even overlap. nothing to draw.

			int screen_x = (int)floor(r.Ulx() + 0.5f);
			int screen_y = (int)floor(r.Uly() + 0.5f);

			int size_x = int(floor(r.Ulx() + r.Xsize() + 0.5f)) - screen_x;
			int size_y = int(floor(r.Uly() + r.Ysize() + 0.5f)) - screen_y;


//			int pic_x = (int)floor(screen_x - upLx);
//			int pic_y = (int)floor(screen_y - upLy);
			int pic_x = (int)floor(screen_x - upLx + 0.5f);
			int pic_y = (int)floor(screen_y - upLy + 0.5f);

/*
LOG(screen_x);
LOG(screen_y);
LOG(size_x);
LOG(size_y);
LOG(pic_x);
LOG(pic_y);
LOG(upLx);
LOG(upLy);
LOG(r.Xsize());
LOG(r.Ysize());
LOG(r.Ulx());
LOG(r.Uly());
*/

			// Sub-pixel stuff.
            if (0 >= size_x || 0 >= size_y) continue; // nothing to draw.


            if (-1 == mDebugLevel || 1 <= mDebugLevel)
                mDebugDraw(pDest,
                           screen_x,
                           screen_y,
                           size_x, size_y);

            else
            {
                // allegro doesn't like bitmaps with zero height-width?

				// Ugly & slow method for transparently blitting parts of bitmaps.
				BITMAP *bmp;
				bmp = create_sub_bitmap(mpGetAllegro(), pic_x, pic_y, size_x, size_y);

				set_alpha_blender();
				draw_trans_sprite(pDest->mpGetAllegro(), bmp, screen_x, screen_y);

				destroy_bitmap(bmp);

				//logerr << "Canvas: sub-sprite copying with alpha not supported. allegro problem." << endl;

				//fblend_trans(GetAllegro(), pDest->GetAllegro(), upLx + 0.5, upLy + 0.5,
				//			 destUpLx + 0.5, destUpLy + 0.5, sizeX + 0.5, sizeY + 0.5);
			}
		}
		
        return;
	}
	
    logerr << "Unknown blending mode." << endl;
    return;
}

	
// pixel edge

void Canvas::mDebugDraw(Canvas* pDest, float upLx, float upLy,
                        float sizeX, float sizeY)
{
    int col;

    // red - copy
    // green - blend
    // blue - unknown

    int transparency = 128; // 0..255: 255 - opaque

    if (ECopy == BlendMode())
        col = makeacol(255, 100, 90, transparency);
    else if (EMulti == BlendMode())
        col = makeacol(30, 255, 30, transparency);
    else
    {
        logerr << "Unknown blending mode." << endl;
        col = makeacol(90, 100, 255, transparency);
    }

    drawing_mode(DRAW_MODE_TRANS, NULL, 0, 0);

    int x = int(upLx + 0.5f);
    int y = int(upLy + 0.5f);
    
    rectfill(pDest->mpGetAllegro(),
	    x, y,
	    x + (int)(upLx + sizeX) - (int)upLx - 1,
	    y + (int)(upLy + sizeY) - (int)upLy - 1,
        col);
    // 255 - opaque.

    solid_mode();
}


void Canvas::Draw(Canvas* pDest, float upLx, float upLy)
{
    if (0 >= Xsize() || 0 >= Ysize()) return;

    if (-1 == mDebugLevel || 11 == mDebugLevel)
        mDebugDraw(pDest, upLx, upLy, Xsize(), Ysize());
        
    else

	if (ECopy == BlendMode()) {

		blit(mpGetAllegro(), pDest->mpGetAllegro(), 0, 0,
			 int(upLx + 0.5f), int(upLy + 0.5f),
			 (int)(upLx + Xsize()) - (int)upLx,
			 (int)(upLy + Ysize()) - (int)upLy);

	} else if (EMulti == BlendMode()) {

		set_alpha_blender();
		draw_trans_sprite(pDest->mpGetAllegro(), mpGetAllegro(), int(upLx + 0.5f), int(upLy + 0.5f));

	}
	else { logerr << "Unknown blending mode." << endl; }
}



/*
void Canvas::Draw(const Canvas* pDest, float upLx, float upLy) const
{
    if (mpPic)
		if (HasUsableAlpha()) {
			set_alpha_blender();
			draw_trans_sprite(pDest->GetAllegro(), GetAllegro(), upLx + 0.5, upLy + 0.5);
		}
	    else
		    blit(GetAllegro(), pDest->GetAllegro(), 0, 0, upLx + 0.5, upLy + 0.5, GetXSize(), GetYSize());
}


void Canvas::Blit(const Canvas* pDest, float upLx, float upLy) const
{
    if (mpPic)
	    blit(GetAllegro(), pDest->GetAllegro(), 0, 0, upLx + 0.5, upLy + 0.5, GetXSize(), GetYSize());
}

// pixel centers

void Canvas::Draw(const Canvas* pDest, float upLx, float upLy,
		           float destUpLx, float destUpLy,
7		           float sizeX, float sizeY) const
{
    if (mpPic)
	    if (HasUsableAlpha()) {
			//logerr << "Canvas: sub-sprite copying with alpha not supported. allegro problem." << endl;
			//exit(2);

			// Ugly & slow method for transparently blitting parts of bitmaps.
			BITMAP *bmp;
			bmp = create_sub_bitmap(GetAllegro(), upLx + 0.5, upLy + 0.5, sizeX + 0.5, sizeY + 0.5);
			set_alpha_blender();
			draw_trans_sprite(pDest->GetAllegro(), bmp, destUpLx + 0.5, destUpLy + 0.5);
			destroy_bitmap(bmp);

			//fblend_trans(GetAllegro(), pDest->GetAllegro(), upLx + 0.5, upLy + 0.5,
			//			 destUpLx + 0.5, destUpLy + 0.5, sizeX + 0.5, sizeY + 0.5);
		} else blit(GetAllegro(), pDest->GetAllegro(), upLx + 0.5, upLy + 0.5,
			        destUpLx + 0.5, destUpLy + 0.5, sizeX + 0.5, sizeY + 0.5);
}
*/
/*

// pixel centers

void Canvas::DrawCenter(const Canvas* pDest, float cX, float cY) const
{
    Draw(pDest, cX - GetXSize() / 2.0f + 0.5f, cY - GetYSize() / 2.0f + 0.5f);
}

*/

// pixel edges

// If down & up are reversed, nothing will get drawn.
void Canvas::Rect(float upLx, float upLy, float downRx, float downRy, unsigned char r, unsigned char g, unsigned char b)
{
    if (upLx < downRx && upLy < downRy) 
	    rect(mpGetAllegro(), int(upLx), int(upLy), int(downRx), int(downRy), makecol(r, g, b));
}


void Canvas::FillRect(float upLx, float upLy, float downRx, float downRy, unsigned char r, unsigned char g, unsigned char b)
{
    if (upLx < downRx && upLy < downRy)
	    rectfill(mpGetAllegro(), int(upLx), int(upLy), int(downRx), int(downRy), makecol(r, g, b));
}


void Canvas::FillRect(RectMesh* pMask, float upLx, float upLy, float downRx, float downRy, unsigned char r, unsigned char g, unsigned char b)
{
    if (NULL == pMask)
    {
        Canvas::FillRect(upLx, upLy, downRx, downRy, r, g, b);
        return;
    }

    if (upLx < downRx && upLy < downRy)
	{
        // *this bounding rectangle.
        Recta th;
        th.Pos(upLx, upLy);
        th.Drx(downRx);
        th.Dry(downRy);

        for (RectMesh::iterator i = pMask->begin(); i != pMask->end(); i++)
        {
			// Some kind of sub-pixel accuracy. hopefully!!

            Recta rt = *i;

            if (!rt.Clip(th)) continue; // They don't even overlap. nothing to draw.

			int screen_x = (int)floor(rt.Ulx() + 0.5f);
			int screen_y = (int)floor(rt.Uly() + 0.5f);

			int screen_d_x = int(floor(rt.Ulx() + rt.Xsize() + 0.5f)) - 1;
			int screen_d_y = int(floor(rt.Uly() + rt.Ysize() + 0.5f)) - 1;

			// Sub-pixel stuff.
            if (screen_d_x <= screen_x || screen_d_y <= screen_y) continue; // nothing to draw.


            if (-1 == mDebugLevel || 1 <= mDebugLevel)
                mDebugDraw(this,
                           screen_x,
                           screen_y,
                           screen_d_x - screen_x, screen_d_x - screen_y);

            else

            rectfill(mpGetAllegro(),
                     screen_x, screen_y,
                     screen_d_x, screen_d_y,
                     makecol(r, g, b));
        }

        return;
	}
}
/*

void Canvas::Circle(float x, float y, float radius, unsigned char r, unsigned char g, unsigned char b)
{
	circle(GetAllegro(), x + 0.5, y + 0.5, radius, makecol(r, g, b));
}


void Canvas::CircleFill(float x, float y, float radius, unsigned char r, unsigned char g, unsigned char b)
{
	circlefill(GetAllegro(), x + 0.5, y + 0.5, radius, makecol(r, g, b));
}
*/
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

/*Canvas::Canvas()
{	

}*/

Canvas::~Canvas()
{
    if (!mAllegParallel && mpPic) destroy_bitmap(mpPic);
}

