// Canvas
//  update: 21.09.03 21:12

// NB!! integer coordinates mark pixel centers! and - on-edge coordinates will be increased.
// ie. lower-right fill-convention I think. when x = 0.4, pixel 0 will be lit. when x = 0.5, pixel
// 1 will be lit. rounding.


#ifndef _CANVAS_H_
#define _CANVAS_H_


struct BITMAP;


#include "string"

//#include "allegro.h"

#include "iGraphics.h"


#include "sys_include.h"


// A Canvas is something you can draw onto. And something you can draw onto another Canvas.


// Everything will be internally 32bpp. Converted on loading, if not.
// (if no alpha, will get a default value of 0).
// NO SUPPORT for <15 bpp images. // old comment. but still no support.
// And also expects to be the only part of the program to read picture files.
class Canvas: public iGraphics
{
public:

	Canvas();
	virtual ~Canvas();



	// When an allegro bitmap is passed to this Canvas, this class will destroy it itself on destruction.
	// So, never pass this method the allegro' screen bitmap :)
    void SetAllegro(BITMAP *pic);

	// On destruction, this class will NOT destroy this assigned bitmap.
	void SetAllegroPointer(BITMAP *pic);


    // Doesn't resize Canvas contents. Only allocates more mem.
	// Previous Canvas contents lost.
    void SetSize(float xSize, float ySize);


	void Clear(int r, int g, int b, int a);


	float Xsize(void);
	float Ysize(void);


	// If loading failed, makes itself a 20x20 green opaque picture.
	void Load(const std::string& rFname);


	// Give NULL for no mask. Simple on-off mask (meaning: no alpha and other fancy stuff).
	// The effect is in that everything that gets drawn to this Canvas is drawn through the mask.
	// Only areas covered by the mask will get drawn.'
	// (random thoughts: this approach basically renders the class-structure useless for 
	// lower-level programming, because every class still has to use allegro internally. So, why
	// not use allegro always? Because I still hope I can port this class-structure somehow 
	// to opengl.)
	//void UseMask(RectMesh* pMesh) { mpMaskMesh = pMesh; }

    void Coverage(RectMesh* pMesh, float x, float y);

	// Makes use of blending mode.
	// upLx is on the pixel edge.
	void Draw(Canvas* pDest, float upLx, float upLy);
	// Only pixels covered by the mask will get drawn.
	//void Draw(Canvas* pDest, RectMesh& rMask, float upLx, float upLy, float meshOffsetUpLx, float meshOffsetUpLy);
	void Draw(Canvas* pDest, RectMesh* pMask, float upLx, float upLy);

	// Pixel centers.
	void Pixel(float x, float y, int r, int g, int b, int a);

/*
	// Should be rather named blit, or copy.
	void Draw(const Canvas* pDest, float upLx, float upLy) const;

	void Blit(const Canvas* pDest, float upLx, float upLy) const;
	
	// Copy part of a Canvas.
	void Draw(const Canvas* pDest, float upLx, float upLy,
		      float destUpLx, float destUpLy,
		      float sizeX, float sizeY) const;

cY) const;


    void Circle(float x, float y, float radius, unsigned char r, unsigned char g, unsigned char b);
    void CircleFill(float x, float y, float radius, unsigned char r, unsigned char g, unsigned char b);
*/

	// Well. char r, g, b is baad. Better make some color class? let it be for now.
    // If down & up are reversed, nothing will get drawn.
    void Rect(float upLx, float upLy, float downRx, float downRy, unsigned char r, unsigned char g, unsigned char b);
    void FillRect(float upLx, float upLy, float downRx, float downRy, unsigned char r, unsigned char g, unsigned char b);
    void FillRect(RectMesh* pMesh, float upLx, float upLy, float downRx, float downRy, unsigned char r, unsigned char g, unsigned char b);


	BITMAP* mpGetAllegro(void);

    void SetDebugLevel(int level);


private:

	BITMAP* mpPic;

	bool mAllegParallel;
	
	int mDebugLevel;

	void mLoad(const char* pFname, int alpha);

    void mDebugDraw(Canvas* pDest, float upLx, float upLy,
                    float sizeX, float sizeY);
};


#include "sys_include_end.h"


#endif

