// Cave.
//  update: 23.11.03 11:04


#include "Cave.h"


#include "allegro.h"

#include "Ball.h"
#include "Canvas.h"
#include "mymath.h"


#include "sys_include.h"


using namespace std;


const float PI = 3.141592654f;

// float to int with rounding
#define FI(f) (int((f) + 0.5f))


/////////////////////////////////// PUBLIC ///////////////////////////////////

// ** START ********************* lifecycle ******************************* //


float cave_damping;


Cave::Cave()
{ 
    cave_damping = get_config_float("Game", "cave_damping", 0.01);
}

Cave::~Cave()
{ 
}

Cave::CavePiece::CavePiece()
{
    x1 = y1 = x2 = y2 = 0;
    color = makecol(100, 100, 100);
}


// ** END *********************** lifecycle ******************************* //

// ** START ********************* operators ******************************* //
// ** END *********************** operators ******************************* //


void Cave::baseStartCoord(float* x, float* y)
{
    assert(m_cave.size() > 0 && x && y);
    
    *x = m_cave[0].x1;
    *y = m_cave[0].y1;
}


void Cave::baseEndCoord(float* x, float* y)
{
    assert(m_cave.size() > 0 && x && y);
    
    *x = m_cave[m_cave.size() - 1].x2;
    *y = m_cave[m_cave.size() - 1].y2;
}


void Cave::draw(Canvas* pDest, float ox, float oy)
{
    assert(pDest);

    BITMAP* canvas = pDest->mpGetAllegro();

    for (vector<CavePiece>::iterator i = m_cave.begin(); i != m_cave.end(); ++i)
    {
        CavePiece& c = *i;
        
        line(canvas, FI(c.x1 + ox), FI(c.y1 + oy), FI(c.x2 + ox), FI(c.y2 + oy), c.color);
    }

    // Draw possible cave-ball collision-points.
    
    if (m_ball)
    {
        Ball &b = *m_ball;
    
        // Calculate the coordinate of nearest spring-point.
        
        int seg_num = m_cave.size();
        for (int i = 0; i < seg_num; ++i)
        {
            CavePiece& p = m_cave[i];
    
            float x, y;
            bool intersects = calc_perpendicular_lineseg_point_intersection
                             (p.x1, p.y1, p.x2, p.y2, b.x, b.y, 
                              &x, &y);
                              
            if (intersects) 
                circlefill(canvas, FI(x + ox), FI(y + oy), 2, makecol(255, 90, 90));
        }
    }
}


void Cave::handleBall(Ball* ball)
{
    assert(ball);
    
    Ball& b = *ball;
    m_ball = ball;

    bool touched_wall = false;
        
    int seg_num = m_cave.size();
    for (int i = 0; i < seg_num; ++i)
    {
        CavePiece& p = m_cave[i];

        float x, y;
        bool intersects = calc_perpendicular_lineseg_point_intersection
                             (p.x1, p.y1, p.x2, p.y2, b.x, b.y, 
                              &x, &y);
        bool touches_endpoint1 = false;
        bool touches_endpoint2 = false;

        contact_x = x;
        contact_y = y;

        // mp1 is inside ball
        if (b.radius * b.radius > (b.x - p.x1) * (b.x - p.x1) + (b.y - p.y1) * (b.y - p.y1))
            touches_endpoint1 = true;
        // mp2 is inside ball
        if (b.radius * b.radius > (b.x - p.x2) * (b.x - p.x2) + (b.y - p.y2) * (b.y - p.y2))
            touches_endpoint2 = true;
                
        if (intersects)
        {
            float ball_caveseg_dist = dist(b.x, b.y, x, y);
            if (ball_caveseg_dist < 0.001f) ball_caveseg_dist = 0.001f;
            if (ball_caveseg_dist < b.radius)
            {
                // reflect the ball's velocity vector
                
                // cavesegment normal
                float unit_normal_x = (b.x - x) / ball_caveseg_dist;
                float unit_normal_y = (b.y - y) / ball_caveseg_dist;

                contact_normal_x = unit_normal_x;
                contact_normal_y = unit_normal_y;
                
                // velocity vector projected to normal. length of it.
                float dot_prod = b.vx * unit_normal_x + b.vy * unit_normal_y;
                // make it only reflect away.
                if (dot_prod < 0)
                {
                    float ball_vx = -b.vx + 2 * dot_prod * unit_normal_x;
                    float ball_vy = -b.vy + 2 * dot_prod * unit_normal_y;
                    
                    b.vx = -ball_vx;
                    b.vy = -ball_vy;
                }
                
                // move ball to the surface.
                b.x += unit_normal_x * (b.radius - ball_caveseg_dist);
                b.y += unit_normal_y * (b.radius - ball_caveseg_dist);
                
                touched_wall = true;
            }
        }
        else if (touches_endpoint1 || (i == seg_num - 1 && touches_endpoint2))
        {
            float ex, ey;
            if (touches_endpoint1) { ex = p.x1; ey = p.y1; } else { ex = p.x2; ey = p.y2; }
            
            float ball_caveseg_dist = dist(b.x, b.y, ex, ey);
            if (ball_caveseg_dist < 0.001f) ball_caveseg_dist = 0.001f;
            float unit_normal_x = (b.x - ex) / ball_caveseg_dist;
            float unit_normal_y = (b.y - ey) / ball_caveseg_dist;

            contact_normal_x = unit_normal_x;
            contact_normal_y = unit_normal_y;

            // velocity vector projected to normal. length of it.
            float dot_prod = b.vx * unit_normal_x + b.vy * unit_normal_y;
            
            // make it only reflect away.
            if (dot_prod < 0)
            {
                //dot_prod = -dot_prod; 
                float ball_vx = b.vx - 2 * dot_prod * unit_normal_x;
                float ball_vy = b.vy - 2 * dot_prod * unit_normal_y;
               
                b.vx = ball_vx;
                b.vy = ball_vy;
            }
            
            // move ball to the surface.
            b.x += unit_normal_x * (b.radius - ball_caveseg_dist);
            b.y += unit_normal_y * (b.radius - ball_caveseg_dist);
            
            touched_wall = true;
        }
    }
    
    // speed limit for the ball..
    
    if (touched_wall)
    {
        const float SPEED_LIMIT = 20;
        
        float speed = sqrt(b.vx * b.vx + b.vy * b.vy);
        if (speed > SPEED_LIMIT)
        {
            b.vx *= SPEED_LIMIT / speed;
            b.vy *= SPEED_LIMIT / speed;
        }
        
        b.vx *= 1 - cave_damping;
        b.vy *= 1 - cave_damping;
        was_contact = true;
    }
    else
        was_contact = false;
}


void Cave::translate(float dx, float dy)
{
    for(vector<CavePiece>::iterator i = m_cave.begin(); i != m_cave.end(); ++i)
    {
        i->x1 += dx;
        i->y1 += dy;
        i->x2 += dx;
        i->y2 += dy;
    }
}


void Cave::rotate(float degree)
{
    assert(m_cave.size() > 0);
    // rotate around the first mass-point. masspoints[0]
    
    float x0 = m_cave[0].x1;
    float y0 = m_cave[0].y1;
    for(vector<CavePiece>::iterator i = m_cave.begin(); i != m_cave.end(); ++i)
    {
        float dx = i->x1 - x0;
        float dy = i->y1 - y0;
        float new_dx = dx * cos(degree * PI / 180.0f) - dy * sin(degree * PI / 180.0f);
        float new_dy = dx * sin(degree * PI / 180.0f) + dy * cos(degree * PI / 180.0f);
        i->x1 = x0 + new_dx;
        i->y1 = y0 + new_dy;

        dx = i->x2 - x0;
        dy = i->y2 - y0;
        new_dx = dx * cos(degree * PI / 180.0f) - dy * sin(degree * PI / 180.0f);
        new_dy = dx * sin(degree * PI / 180.0f) + dy * cos(degree * PI / 180.0f);
        i->x2 = x0 + new_dx;
        i->y2 = y0 + new_dy;
    }
}


void Cave::generate(int segnum, float seglen, float amplitude, int smooth)
{
    //const int seg_num = 200;
    int seg_num = segnum;
    
    CavePiece dummy;
    for (int i = 0; i < seg_num; i++) m_cave.push_back(dummy);
    
    int n = 0;

    //int y_amplitude = 180;
    int y_amplitude = FI(amplitude);
    float y_pos = 0;
    float prev_x = 0;
    float prev_y = rand() % y_amplitude + y_pos;

    float skew_y = 0;
    
    // generate random heights.    
    for (vector<CavePiece>::iterator i = m_cave.begin(); i != m_cave.end(); ++i)
    {
        CavePiece& c = *i;
        
        // segment length     
        c.x1 = prev_x;
        c.y1 = prev_y;
        
        c.x2 = prev_x = (n + 1) * seglen;
        c.y2 = prev_y = rand() % y_amplitude + y_pos + skew_y;
        
        n++;
    }
    
    
    float buf[seg_num + 1];
    
    for (int n = 0; n < smooth; ++n)
    {
        // smoothen the height-field.
        for (int i = 1; i < seg_num; ++i)
        {
            CavePiece& p1 = m_cave[i - 1];
            CavePiece& p2 = m_cave[i];
    
            float mid_y = (p1.y1 + p2.y2) / 2;
            buf[i] = p1.y2 + (mid_y - p1.y2) / 5;
        }
        // smoothed field from buf to m_cave
        for (int i = 1; i < seg_num; ++i)
        {
            CavePiece& p1 = m_cave[i - 1];
            CavePiece& p2 = m_cave[i];
    
            p1.y2 = p2.y1 = buf[i];
        }
    }
    
    loginf << "generated a cave-segment with " << m_cave.size() << " lines." << endl;
}


///////////////////////////////// PROTECTED //////////////////////////////////

////////////////////////////////// PRIVATE ///////////////////////////////////


