// Jelly.
//  update: 15.06.03 15:31


#include "Jelly.h"

#include "allegro.h"

#include "Canvas.h"
#include "Ball.h"
#include "mymath.h"


#include "sys_include.h"


using namespace std;


const float PI = 3.141592654f;

// float to int with rounding
#define FI(f) (int((f) + 0.5f))


/////////////////////////////////// PUBLIC ///////////////////////////////////

// ** START ********************* lifecycle ******************************* //


MassPoint::MassPoint()
{ 
    x = y = vx = vy = fx = fy = 0;
    m = 1;
    anchor = false; color = makecol(255, 200, 166); visible = true;
}


Spring::Spring()
{
    l = 0; p1 = p2 = NULL; k = 0.8;
    color = makecol(155, 155, 100);
    visible = true;
    outer_edge = false;
    damping = 0.001;  
}


Jelly::Jelly()
{
    damping = 0.01; m = 1; timestep = 0.1;
    ball = NULL;
    m_base_start = m_base_end = NULL;
    contacts = 0;

    jelly_internals_visible = get_config_int("Game", "jelly_internals_visible", 1);
    jelly_base_visible = get_config_int("Game", "jelly_base_visible", 1);
    jelly_points_visible = get_config_int("Game", "jelly_points_visible", 1);  
}


Jelly::~Jelly()
{
}


// ** END *********************** lifecycle ******************************* //

// ** START ********************* operators ******************************* //
// ** END *********************** operators ******************************* //


void Jelly::baseStartCoord(float* x, float* y)
{
    assert(m_base_start && x && y);
    
    *x = m_base_start->x;
    *y = m_base_start->y;
}


void Jelly::baseEndCoord(float* x, float* y)
{
    assert(m_base_end && x && y);
    
    *x = m_base_end->x;
    *y = m_base_end->y;
}


float dist(MassPoint* p1, MassPoint* p2)
{
    float dx = p2->x - p1->x; float dy = p2->y - p1->y;
    return sqrt(dx*dx + dy*dy);
}


void Jelly::translate(float dx, float dy)
{
    for(vector<MassPoint>::iterator i = masspoints.begin(); i != masspoints.end(); ++i)
    {
        i->x += dx;
        i->y += dy;
    }
}


void Jelly::rotate(float degree)
{
    assert(masspoints.size() > 0);
    // rotate around the first mass-point. masspoints[0]
    
    float x0 = masspoints[0].x;
    float y0 = masspoints[0].y;
    for(vector<MassPoint>::iterator i = masspoints.begin(); i != masspoints.end(); ++i)
    {
        float dx = i->x - x0;
        float dy = i->y - y0;
        float new_dx = dx * cos(degree * PI / 180.0f) - dy * sin(degree * PI / 180.0f);
        float new_dy = dx * sin(degree * PI / 180.0f) + dy * cos(degree * PI / 180.0f);
        i->x = x0 + new_dx;
        i->y = y0 + new_dy;
    }
}


void Jelly::generateBlob(float baseLen, float height, int topsurfaceNum)
{
    springs.clear(); masspoints.clear();
    
    // create the masspoints.
    MassPoint mp;
    for (int i = 0; i < (2 * topsurfaceNum); i++) 
    {
      mp.y = i * 5; masspoints.push_back(mp);
    }
 
    // init mass-points

    // create the base

    for (int i = 0; i <= topsurfaceNum; i++)
    {
        MassPoint& mp = masspoints[i];
        mp.m = 1;
        mp.x = baseLen / topsurfaceNum * i;
        mp.y = 0;
        mp.anchor = true;
    }
    
    m_base_start = &masspoints[0];
    m_base_end = &masspoints[topsurfaceNum];
    
    int cover_index = topsurfaceNum + 1;
    
    // create the cover
    
    for (int i = 1; i < topsurfaceNum; i++)
    {
        float x = i * (baseLen / topsurfaceNum);
        float xx = x - baseLen / 2;
        float r = baseLen / 2 - 16;
        float y;
        if (xx < r) y = -sqrt(r * r - xx * xx);
        else y = -20;
        
/*
        float angle = i * (180 / (topsurfaceNum));
        
        float x = -baseLen / 2 * cos(angle * PI / 180) + baseLen / 2;
        float y = -height / 2 * sin(angle * PI / 180);
*/      
        MassPoint& mp = masspoints[cover_index + i - 1];
        mp.m = 1;
        mp.x = x;
        mp.y = y;
        
//        masspoints[i].x = x;
        //if (!(yi == 0 || yi == yNum || xi == 0 || xi == xNum)) mp.visible = false;
        //mp.color = makecol(255, 200, 166);
    }
    
    // create and init springs
    for (int i = 0; i < topsurfaceNum; i++)
    {
        // base-springs
        Spring s;
        s.p1 = &masspoints[i];
        s.p2 = &masspoints[i + 1];
        s.l = dist(s.p1, s.p2);
        //s.k = 0.01;
        //s.color = makecol(255, 255, 200);
        //if (yi != 0 && yi != yNum) s.visible = false;
        s.outer_edge = true;
        springs.push_back(s);
    }

    for (int i = 0; i < topsurfaceNum - 2; i++)
    {
        // cover-springs
        Spring s;
        s.p1 = &masspoints[i + cover_index + 1];
        s.p2 = &masspoints[i + cover_index];
        s.l = dist(s.p1, s.p2);
        //s.k = 0.01;
        //s.color = makecol(255, 255, 200);
        //if (yi != 0 && yi != yNum) s.visible = false;
        s.outer_edge = true;
        springs.push_back(s);
    }

    {
        // cover
        Spring s;
        s.p1 = &masspoints[0];
        s.p2 = &masspoints[cover_index];
        s.l = dist(s.p1, s.p2);
        //s.k = 0.01;
        //s.color = makecol(255, 255, 200);
        //if (yi != 0 && yi != yNum) s.visible = false;
        s.outer_edge = true;
        springs.push_back(s);
    }
    {
        // cover
        Spring s;
        s.p1 = &masspoints[topsurfaceNum * 2 - 1];
        s.p2 = &masspoints[cover_index - 1];
        s.l = dist(s.p1, s.p2);
        //s.k = 0.01;
        //s.color = makecol(255, 255, 200);
        //if (yi != 0 && yi != yNum) s.visible = false;
        s.outer_edge = true;
        springs.push_back(s);
    }


    // internal springs.. flower
    for (int i = 0; i < topsurfaceNum - 1; i++)
    {
        Spring s;
        s.p1 = &masspoints[cover_index + i];
        s.p2 = &masspoints[(topsurfaceNum) >> 1];
        s.l = dist(s.p1, s.p2);
        //s.k = 0.01;
        //s.color = makecol(255, 255, 200);
        //if (yi != 0 && yi != yNum) s.visible = false;
        //s.visible = false;
        springs.push_back(s);
    }

    // internal springs.. vertical
    for (int i = 1; i < topsurfaceNum; i++)
    {
        Spring s;
        s.p1 = &masspoints[i];
        s.p2 = &masspoints[cover_index + i - 1];
        s.l = dist(s.p1, s.p2);
        //s.k = 0.01;
        //s.color = makecol(255, 255, 200);
        //if (yi != 0 && yi != yNum) s.visible = false;
        //s.visible = false;
        springs.push_back(s);
    }
    
    loginf << "generated " << springs.size() << " springs and " 
           << masspoints.size() << " mass-points." << endl;
}


void Jelly::generateRectangular(float springLen, int xNum, int yNum)
{
    springs.clear(); masspoints.clear();
    
    // create the masspoints.
    MassPoint mp;
    for (int i = 0; i < (xNum + 1) * (yNum + 1); i++) masspoints.push_back(mp);
    
    // init mass-points
    for (int yi = 0; yi <= yNum; yi++)
        for (int xi = 0; xi <= xNum; xi++)
        {
            int i = xi + yi * (xNum + 1);
            MassPoint& mp = masspoints[i];
            mp.m = 1;
            mp.x = xi * springLen;
            mp.y = yi * springLen;
            if (yi == yNum)
            {
                if (xi == 0) m_base_start = &mp;
                if (xi == xNum) m_base_end = &mp;
            
                mp.anchor = true;
            }
            
            //if (!(yi == 0 || yi == yNum || xi == 0 || xi == xNum)) mp.visible = false;
            //mp.color = makecol(255, 200, 166);
        }
    
    
    // create and init springs
    for (int yi = 0; yi <= yNum; yi++)
    {
        for (int xi = 0; xi <= xNum; xi++)
        {
            int i = xi + yi * (xNum + 1);
            if (xi < xNum) {
                // horizontal springs
                Spring s;
                s.p1 = &masspoints[i];
                s.p2 = &masspoints[i + 1];
                s.l = dist(s.p1, s.p2);
                //s.k = 0.01;
                //s.color = makecol(255, 255, 200);
                //if (yi != 0 && yi != yNum) s.visible = false;
                if (yi == 0 || yi == yNum) s.outer_edge = true;
                springs.push_back(s);
            }
            if (yi < yNum) {
                // vertical springs
                Spring s;
                s.p1 = &masspoints[i];
                s.p2 = &masspoints[i + xNum + 1];
                s.l = dist(s.p1, s.p2);
                //s.k = 0.01;
                //s.color = makecol(255, 255, 200);
                //if (xi != 0 && xi != xNum) s.visible = false;
                if (xi == 0 || xi == xNum) s.outer_edge = true;
                springs.push_back(s);
            }
            if (xi < xNum && yi < yNum) {
                // diagonal springs 1
                Spring s;
                s.p1 = &masspoints[i];
                s.p2 = &masspoints[i + xNum + 1 + 1];
                s.l = dist(s.p1, s.p2);
                //s.k = 0.01;
                //s.color = makecol(255, 255, 200);
                s.visible = false;
                springs.push_back(s);
            }
            if (xi < xNum && yi < yNum) {
                // diagonal springs 2
                Spring s;
                s.p1 = &masspoints[i + 1];
                s.p2 = &masspoints[i + xNum + 1];
                s.l = dist(s.p1, s.p2);
                //s.k = 0.01;
                //s.color = makecol(255, 255, 200);
                s.visible = false;
                springs.push_back(s);
            }
/*            
            if (xi < xNum - 1 && yi < yNum - 1) {
                // diagonal springs 3, over2 horizontal...
                Spring s;
                s.p1 = &masspoints[i + 1];
                s.p2 = &masspoints[i + xNum + 1 + xNum + 1];
                s.l = dist(s.p1, s.p2);
                //s.k = 0.01;
                s.color = makecol(255, 11, 200);
                s.visible = true;
                springs.push_back(s);
            }
            if (xi < xNum - 1 && yi < yNum - 1) {
                // diagonal springs 3, over2 horizontal...
                Spring s;
                s.p1 = &masspoints[i];
                s.p2 = &masspoints[i + xNum + 1 + xNum + 1 + 1];
                s.l = dist(s.p1, s.p2);
                //s.k = 0.01;
                s.color = makecol(255, 11, 200);
                s.visible = true;
                springs.push_back(s);
            }
*/            
        }
    }
    loginf << "generated " << springs.size() << " springs and " 
           << masspoints.size() << " mass-points." << endl;
}


bool is_stationary_spring(Spring* s)
{
    assert(s);
    assert(s->p1);
    assert(s->p2);
    if (s->p1->anchor && s->p2->anchor) return true;
    else return false;
}


void Jelly::handleBall()
{
    // collision response tuning..
    const float FORCE_TO_JELLY_MULTIPLIER = 2;
    // actually - it's k. stiffness. like there is a spring formed between the
    // jelly and the ball.
    const float FORCE_TO_BALL = 3;

    was_contact = false;
    
    if (ball)
    {
        Ball &b = *ball;
    
        // Calculate the coordinate of nearest spring-point.
        
        for(vector<Spring>::iterator i = springs.begin(); i != springs.end(); ++i)
        {
            if (i->outer_edge)
            {
                MassPoint& mp1 = *(i->p1);
                MassPoint& mp2 = *(i->p2);
        
                float x, y;
                bool intersects = calc_perpendicular_lineseg_point_intersection
                                     (mp1.x, mp1.y, mp2.x, mp2.y, b.x, b.y, 
                                      &x, &y);

                bool touches_endpoint1 = false;
                bool touches_endpoint2 = false;

                contact_x = x;
                contact_y = y;

                // mp1 is inside ball
                if (b.radius * b.radius > (b.x - mp1.x) * (b.x - mp1.x) + (b.y - mp1.y) * (b.y - mp1.y))
                    touches_endpoint1 = true;
                // mp2 is inside ball
                if (b.radius * b.radius > (b.x - mp2.x) * (b.x - mp2.x) + (b.y - mp2.y) * (b.y - mp2.y))
                    touches_endpoint2 = true;
                                                                              
                if (intersects)
                {
                    // if ball really intersects with this point?
                    // point inside ball.
                    float ball_spring_dist = dist(b.x, b.y, x, y);
                    if (ball_spring_dist < 0.001f) ball_spring_dist = 0.001f;
                    if (ball_spring_dist < b.radius)
                    {
                        float d = dist(&mp1, &mp2);
                        //float spring_len = d;
                        //float unit_normal_x = (mp1.y - mp2.y) / spring_len;
                        //float unit_normal_y = (mp2.x - mp1.x) / spring_len;
                        float unit_normal_x = (b.x - x) / ball_spring_dist;
                        float unit_normal_y = (b.y - y) / ball_spring_dist;
                        
                        contact_normal_x = unit_normal_x;
                        contact_normal_y = unit_normal_y;
                
                        //float response_force = 3;
                        //float response_multiplier = ball_spring_dist / b.radius;

                        float response_force = -FORCE_TO_BALL * (ball_spring_dist - b.radius);
                        
                        /*
                        // push the ball away
                        b.x += unit_normal_x * ball_spring_dist * 3;
                        b.y += unit_normal_y * ball_spring_dist * 3;
                        
                        // push the spring away
                        mp1.x += -unit_normal_x * ball_spring_dist;
                        mp1.y += -unit_normal_y * ball_spring_dist;
                        mp2.x += -unit_normal_x * ball_spring_dist;
                        mp2.y += -unit_normal_y * ball_spring_dist;
                        */                        
                        
                        // push the ball away
                        b.fx += unit_normal_x * response_force;
                        b.fy += unit_normal_y * response_force;
                        
                        // push the spring away
                        float dist_xy_mp1 = dist(x, y, mp1.x, mp1.y);
                        if (dist_xy_mp1 > 0.001f)
                        {
                            mp2.fx += -unit_normal_x * response_force * (dist_xy_mp1 / d) * FORCE_TO_JELLY_MULTIPLIER;
                            mp2.fy += -unit_normal_y * response_force * (dist_xy_mp1 / d) * FORCE_TO_JELLY_MULTIPLIER;
                        }
                        if ((d - dist_xy_mp1) > 0.001f)
                        {
                            mp1.fx += -unit_normal_x * response_force * ((d - dist_xy_mp1) / d) * FORCE_TO_JELLY_MULTIPLIER;
                            mp1.fy += -unit_normal_y * response_force * ((d - dist_xy_mp1) / d) * FORCE_TO_JELLY_MULTIPLIER;
                        }                     
                        
                        contacts++;
                        was_contact = true;
                    }
                }
                else if (touches_endpoint1 || touches_endpoint2)
                {
                    MassPoint* mp;
                    if (touches_endpoint1) mp = &mp1; else mp = &mp2;
                    
                    float ball_springendpoint_dist = dist(b.x, b.y, mp->x, mp->y);
                    if (ball_springendpoint_dist < 0.001f) ball_springendpoint_dist = 0.001f;
                    float unit_normal_x = (b.x - mp->x) / ball_springendpoint_dist;
                    float unit_normal_y = (b.y - mp->y) / ball_springendpoint_dist;

                    contact_normal_x = unit_normal_x;
                    contact_normal_y = unit_normal_y;
                        
                    // div 2, cause there are 2 springs in this collision.
                    float response_force = -FORCE_TO_BALL * (ball_springendpoint_dist - b.radius) / 2;
                    
                    // push the ball away
                    b.fx += unit_normal_x * response_force;
                    b.fy += unit_normal_y * response_force;
                    
                    // push the spring away
                    mp->fx += -unit_normal_x * response_force * FORCE_TO_JELLY_MULTIPLIER;
                    mp->fy += -unit_normal_y * response_force * FORCE_TO_JELLY_MULTIPLIER;

                    contacts++;
                    was_contact = true;
                }
            }
        }
    
    
        /*
        // this is in a very-very wrong place!
        
        float dummy = timestep / b.m;
        b.vx += dummy * b.fx;
        b.vy += dummy * b.fy;
        b.vx *= 1 - b.damping;
        b.vy *= 1 - b.damping;
    
        // s = v0*t + a*t*t/2
        //mp.x += mp.vx * timestep + dummy * mp.fx * timestep / 2;
        //mp.y += mp.vy * timestep + dummy * mp.fy * timestep / 2;
        b.x += b.vx * timestep + b.fx / b.m * timestep * timestep / 2;
        b.y += b.vy * timestep + b.fy / b.m * timestep * timestep / 2;

        b.fx = b.fy = 0;*/
    }    
    // -------------------------
}


void Jelly::update()
{
//    for (int times = 0; times < 5; times++)
    {
        handleBall();
   
        // calculate all spring-forces and save them in masspoints for later use.   
        for(vector<Spring>::iterator i = springs.begin(); i != springs.end(); ++i)
        {
            // euler integrator. whatever that means..
    
            MassPoint& mp1 = *(i->p1);
            MassPoint& mp2 = *(i->p2);
    
            float d = dist(&mp1, &mp2);
            if (0 == d) d = 0.1; // don't know if this helps. unstability comes anyway.
            
            // f = -k * delta_x
            float f = i->k * (d - i->l);
            
            // calc force vector xy components using f & d & positions. add forces.
            float fx = (mp2.x - mp1.x) / d * f; 
            float fy = (mp2.y - mp1.y) / d * f; 
            mp1.fx += fx;
            mp2.fx -= fx;
            mp1.fy += fy;
            mp2.fy -= fy;
        }
        
        assert(springs.size() > 0);
        float damping = springs[0].damping;
            
        // move masspoints according to the forces acting on each one of them.
        for(vector<MassPoint>::iterator i = masspoints.begin(); i != masspoints.end(); ++i)
        {
            MassPoint& mp = *i;
            if (!mp.anchor)
            {
                // v = v0 + a*t
                // f = m*a
                // v = v0 + timestep * f / m
                float dummy = timestep / mp.m;
                mp.vx += dummy * mp.fx;
                mp.vy += dummy * mp.fy;
                mp.vx *= 1 - damping;
                mp.vy *= 1 - damping;

                // s = v0*t + a*t*t/2
                //mp.x += mp.vx * timestep + dummy * mp.fx * timestep / 2;
                //mp.y += mp.vy * timestep + dummy * mp.fy * timestep / 2;
                mp.x += mp.vx * timestep + mp.fx / mp.m * timestep * timestep / 2;
                mp.y += mp.vy * timestep + mp.fy / mp.m * timestep * timestep / 2;
            }
        }
    
        // clear all forces.
        for(vector<MassPoint>::iterator i = masspoints.begin(); i != masspoints.end(); ++i)
        {
            i->fx = i->fy = 0;
        }
    }
}


void Jelly::draw(Canvas* pDest, float ox, float oy)
{
    assert(pDest);

    BITMAP* canvas = pDest->mpGetAllegro();

    for(vector<Spring>::iterator i = springs.begin(); i != springs.end(); ++i)
    {
        int color = i->color;
        bool anchor = false;
        if (is_stationary_spring(&(*i))) { anchor = true; color = makecol(0, 0, 0); }
        /* if (i->outer_edge)
        {
            line(canvas, FI(i->p1->x + ox), FI(i->p1->y + oy), 
                         FI(i->p2->x + ox), FI(i->p2->y + oy), color);
        }*/
        
        if (i->visible && (jelly_internals_visible || i->outer_edge) && !anchor)
        {
            line(canvas, FI(i->p1->x + ox), FI(i->p1->y + oy), 
                         FI(i->p2->x + ox), FI(i->p2->y + oy), color);
        }
        else if (anchor && jelly_base_visible)
        {
            line(canvas, FI(i->p1->x + ox), FI(i->p1->y + oy), 
                         FI(i->p2->x + ox), FI(i->p2->y + oy), color);
        }
    }

    for(vector<MassPoint>::iterator i = masspoints.begin(); i != masspoints.end(); ++i)
    {
        int color = i->color;
        if (i->anchor) color = makecol(0, 0, 0);
        if (i->visible && jelly_points_visible)
            circlefill(canvas, FI(i->x + ox), FI(i->y + oy), 3, color);

        //if (masspoints.begin() == i)
        //    circlefill(canvas, FI(i->x + ox), FI(i->y + oy), 3, makecol(0, 255, 0));
    }



    // Draw possible jelly-ball collision-points.
    
    if (ball)
    {
        Ball &b = *ball;
    
        // Calculate the coordinate of nearest spring-point.
        
        for(vector<Spring>::iterator i = springs.begin(); i != springs.end(); ++i)
        {
            if (i->outer_edge)
            {
                MassPoint& mp1 = *(i->p1);
                MassPoint& mp2 = *(i->p2);

                float x, y;
                bool intersects = calc_perpendicular_lineseg_point_intersection
                                     (mp1.x, mp1.y, mp2.x, mp2.y, b.x, b.y, 
                                      &x, &y);
                if (intersects) 
                    circlefill(canvas, FI(x + ox), FI(y + oy), 2, makecol(255, 90, 90));
            }
        }

        // -----------------------------
    }
}


void Jelly::generateTriangular(float springLen, int gridNumX, int gridNumY)
{
//    for(vector<Spring>::iterator i = springs.begin(); i != springs.end(); ++i)
    {
    }
}


///////////////////////////////// PROTECTED //////////////////////////////////

////////////////////////////////// PRIVATE ///////////////////////////////////


