// Mouse.
//  update: 15.06.03 15:31


#include "Mouse.h"


#include "allegro.h"


#include "sys_include.h"

using namespace std;


/////////////////////////////////// PUBLIC ///////////////////////////////////

// ** START ********************* lifecycle ******************************* //


Mouse::Mouse()
{
    int mouse_code = install_mouse();
	assert(-1 != mouse_code);
	
	loginf << "\nMouse has " << mouse_code << " buttons.\n" << endl;

    // if allegro has no mouse sprite
	if (NULL == mouse_sprite)
    {
		logmsg << "Allegro's mouse_sprite not there? "
                  "No problem, installing a replacement." << endl;
		mMouseSprite.SetAllegro(create_bitmap(8, 8));
		mLocalMouseSprite = true;
		mMouseSprite.Clear(0, 100, 50, 255);
	}
	else // allegro has a mouse sprite
	{
        mLocalMouseSprite = false;
        //mMouseSprite.BlendMode(Canvas::EMulti);
        mMouseSprite.SetAllegroPointer(mouse_sprite);
	}
	
    mMouseBack.SetSize(mMouseSprite.Xsize(), mMouseSprite.Ysize());
	mOutside = true;

    mX = mY = 0;
	mXdif = mYdif = 0;
    mBackX = mBackY = -1;
    
    Poll();
    
	mPrevX = mX; mPrevY = mY;
	mPrevButton = mButton;

	set_mouse_speed(1, 1);
}


Mouse::~Mouse()
{
}


// ** END *********************** lifecycle ******************************* //

// ** START ********************* operators ******************************* //
// ** END *********************** operators ******************************* //


void Mouse::Poll()
{
	mPrevX = mX; mPrevY = mY;
	mPrevButton = mButton;

	poll_mouse();
	mX = mouse_x; mY = mouse_y;
	mButton = mouse_b;

    get_mouse_mickeys(&mMouseMickyX, &mMouseMickyY);
}


void Mouse::Draw(Canvas* pDest, float upLx, float upLy)
{
    //loginf << mOutside << std::endl;
    int mx = mMouseMickyX, my = mMouseMickyY;

	// filter out small movements that are not reflected on the cursor movement. (ie. mouse
	// moving, cursor not.). not perfect though. but it's good enough. only rare flickering.
	// and no flickering, when set_mouse_speed(1, 1).
	if (1 == mx || -1 == mx || 1 == my || -1 == my) { mx = my = 0; }

    //loginf << " " << mx << " " << my << std::endl;
    //loginf << " " << mouse_x << " " << mouse_y << std::endl;
	// In windowed mode, when mouse goes outside the window, mickeys change,
	// but mouse_x,y don't. But mouse cursor needs to be erased :)
	if (mBackX == mX && mBackY == mY)
    {
		// The cursor has moved! But not in our window. Erase...
		if (0 != mx || 0 != my)
        {
			mOutside = true;
			return;
		}
		if (mOutside) return;
	}

    mOutside = false;

	mBackX = mX; mBackY = mY;

    // Copy the unmodified mouse-pointer background to an internal buffer.
    pDest->Draw(&mMouseBack, -(mBackX + mXdif + upLx + 1), -(mBackY + mYdif + upLy + 1));
    
    // Copy the mouse-pointer to screen.
    if (mLocalMouseSprite)
        mMouseSprite.Draw(pDest, mBackX + mXdif + upLx, mBackY + mYdif + upLy);
    else
        draw_sprite(pDest->mpGetAllegro(), mMouseSprite.mpGetAllegro(),
                    mBackX + mXdif + (int)(upLx + 0.5f),
                    mBackY + mYdif + (int)(upLy + 0.5f));
}


void Mouse::Erase(Canvas* pDest, float upLx, float upLy)
{
	static bool erased = false;
    
    // When outside, erase only once. That's all.

    // mBack == -1 only when Draw not yet called
	if (mOutside && !erased && mBackX != -1)
	{
		erased = true;
        mMouseBack.Draw(pDest, mBackX + mXdif + upLx, mBackY + mYdif + upLy);
    }

    if (!mOutside)
    {
		erased = false;
        mMouseBack.Draw(pDest, mBackX + mXdif + upLx, mBackY + mYdif + upLy);
    }
}


void Mouse::Coverage(RectMesh* pMesh, float x, float y)
{
    mMouseSprite.Coverage(pMesh, x + mX + mXdif, y + mY + mYdif);
}


///////////////////////////////// PROTECTED //////////////////////////////////

////////////////////////////////// PRIVATE ///////////////////////////////////

