// PulseGenerator
//  update: 22.11.03 15:35


#include "PulseGenerator.h"

#include "PulseGeneratorConnection.h"


#include "iObserver.h"

#include "sys_include.h"


using namespace std;


/////////////////////////////////// PUBLIC ///////////////////////////////////

// ** START ********************* lifecycle ******************************* //


PulseGenerator::PulseGenerator(): mObsCounter(0), mAttachCounter(0)
{
    mFreq = -1;
    mbPause = false;
    
    mLastErased = 0;
}


PulseGenerator::~PulseGenerator()
{
    // Really not a strict requirement, but helps to keep things clean.
    // Assures that observers don't live longer than the observable object.
    assert(0 == mObservers.size());
}


// ** END *********************** lifecycle ******************************* //

// ** START ********************* operators ******************************* //
// ** END *********************** operators ******************************* //


void PulseGenerator::Freq(float f)
{
    mFreq = f;
}


float PulseGenerator::Freq()
{
    return mFreq;
}


void PulseGenerator::DoPause(bool b)
{
    mbPause = b;
}


bool PulseGenerator::DoPause()
{
    return mbPause;
}


// returns n2 - n1 if n2 > n1
// else returns !n2 - n1 (assumes that n2 has overflowed and is now behind n1)
uint64 get_distance(uint64 n1, uint64 n2)
{
	// !n1 is the same as 1111...(64 ones in binary) minus n1.
	if (n1 > n2) return !n1 + n2;
	    else return n2 - n1;
}


// at the same time while processing the list, notified objects can remove
// elements from the same list. BUMMER.
// frequency is always a bit-bit lower than requested. but it's simple to
// make it this way. no time to fix :) TODO: <
// but it has to be in sync, at least with the last frame! or shit will happen.
void PulseGenerator::Notify(void)
{
    mObserversType::iterator iter, iter_next;

//and_over:

    mLastErased = 0;
    //mDetachAllCalled = false;

	// TODO: size? wrong size? what? look the log file...
    //loginf << "size: " << mObservers.size() << endl;
	//TODO: log update times and compare. see, if they are in sync.

	iter = mObservers.begin();
	while(iter != mObservers.end())
    {
		iter_next = iter;
		iter_next++;

		int64 dist = get_distance((*iter).last_counts, mTicker.GetTimerCount());
        int64 calldist = int64(1 / (*iter).freq *
                               (int64)mTicker.GetTimerFrequency());

		// async... (the number 3 there) // it means ticks it should take times 3 >= ticks it took
		// wtf does that comment mean? ^
		if (dist > calldist * 3  || !(*iter).firstly_pause || (*iter).freq < 0)
        {
            if ((*iter).freq > 0 && (*iter).firstly_pause) loginf << "updates skipped."<< endl;
			(*iter).last_counts = mTicker.GetTimerCount();
            (*iter).observer->update((*iter).connection);
			(*iter).firstly_pause = true;
		}
        else if (dist > calldist)
        { // to get back to sync. skip what have been skipped.
			(*iter).last_counts += calldist;
            (*iter).observer->update((*iter).connection);
			(*iter).firstly_pause = true;
        }


		// Updated observer decided to destruct itself, possibly taking half
        // of the observers list with itself.
        //if (mDetachAllCalled) goto and_over;

		////if (iter_next == mObservers.end()) break;
		if (iter == mLastErased) iter = iter_next; else iter++;
	}
}


// TODO: attach to the end or to the beginning. meaning when to first call the observer.
SmartPtr<iConnection> PulseGenerator::Attach(iObserver* obs)
{
	assert(0.0f != mFreq); // or division by zero somewhere

/*
        // Make sure the obs is not already registered.
        mObserversType::iterator i;
    	for(i = mObservers.begin(); i != mObservers.end(); i++)
    	{
            if ((*i).observer == obs) assert(false);
        }
*/

	mObsCounter++;
	mAttachCounter++;
	if (0 == mAttachCounter) logmsg << "WOW! 32 bits is not enough! from this"
                         << " point on, there can be random crashes." << endl;

    // create connection. connect.
    PulseGeneratorConnection* conn = new PulseGeneratorConnection;
    conn->mConnect(this, mAttachCounter);

    // now fill in the connection slot info.
	mTimeStruct time_struct;
	time_struct.freq = mFreq;
	time_struct.observer = obs;
    time_struct.obs_code = mAttachCounter;
    time_struct.last_counts = mTicker.GetTimerCount();
    time_struct.firstly_pause = mbPause;
    time_struct.connection = conn;

    mObservers.push_back(time_struct);


    return SmartPtr<iConnection>(conn);
}


///////////////////////////////// PROTECTED //////////////////////////////////

////////////////////////////////// PRIVATE ///////////////////////////////////


// Only PulseGeneratorConnection should use it.
void PulseGenerator::mDetach(int code)
{
    mObserversType::iterator i;

    bool erased = false;

    // Find and remove the proper entry.
	for(i = mObservers.begin(); i != mObservers.end(); i++)
	{
        if ((*i).obs_code == code)
        {
            mObservers.erase(i);
            erased = true;
            mLastErased = i;
            break;
        }
    }

/*
        // Make sure the pObs is not twice, or more registered.
    	for(i = mObservers.begin(); i != mObservers.end(); i++)
    	{
            if ((*i).obs_code == code) assert(false);
        }
*/

    // Element not found. Element MUST be found. To keep things clean.
    assert(erased == true);
}


