// Simple rectangle class, by ETM.
//  update: 04.11.2002


#ifndef _RECTMESH_H_
#define _RECTMESH_H_


#include <list>

#include "Recta.h"


class Canvas;


#include "sys_include.h"


//////////////////////////////////////////////////////////////////////////////


class RectMesh//: public std::list<Recta>
{
public:

    typedef std::list<Recta>::iterator iterator;

	inline iterator begin() { return mRects.begin(); }
	inline iterator end() { return mRects.end(); }

    // LIFECYCLE -------------------------------------------------------------

    RectMesh();

    // OPERATORS -------------------------------------------------------------

    // OPERATIONS ------------------------------------------------------------


	// Adds a rectangle to this mesh. Splits, deletes, reshapes rectangles as necessary, so no
	// area is covered more than once.
	// Don't forget! Rects are completely around pictures (when working with pictures). So
	// a picture with coordinates 0, 0, size 10, 5, will have a bounding rectangle with
	// coordinates -0.5, -0.5, size 10, 5.
    void AddRect(Recta& rRect);
    void AddRect(float uLx, float uLy, float xSize, float ySize);
    void AddRectPos(float centerX, float sizeX, float centerY, float sizeY);


	// First, translate the given mesh. Then clip & make it current (*this). Ugly. Really.
	// rMesh stays intact.
	void SetClipped(RectMesh& rMesh, Recta& rClipRect, float x, float y);
	// First clip, then translate.
	void Clip(Recta& rClipRect, float x, float y);
    // Just clip. A bit faster than the other Clip method, cause it skips the translation part.
    void Clip(Recta& rClipRect);
/*
	// well-well. maybe it would be good to have this one day too! but it's so simple to do now.
    //void DelRect(Rectangle2d& rRect);
*/
    // cut out other mesh. non-overlapping regions stay.
	void Cut(RectMesh& rMesh, float x, float y);
/*
    // no comments, ok :)
	void Join(Rect2dMesh& rMesh);
*/
	// positioned mesh..
	void Join(RectMesh& rMesh, float x, float y);

    void Translate(float x, float y);

	void Clear(void) { mRects.clear(); /*mCurNumRects = 0;*/ }
/*
	// Copies the area covered with rectangles. No area copied twice or more. 
    void Blit(Picture& rSource, Picture* pDest);
	// with offset. of rSource. (rSource coordinates on destination)
    void Blit(Picture& rSource, Picture* pDest, float x, float y); 
    void Blit2(Picture& rSource, Picture* pDest, float x, float y, float x2, float y2); 

    void DebugDraw(Picture* pDest);*/
    void DebugDraw(Canvas* pDest, float x, float y);
/*
    // ACCESS ----------------------------------------------------------------

	void SetDebugLevel(int level) { mDebugLevel = level; }

    // INQUIRY ---------------------------------------------------------------

	bool IsEmpty(void) { if (0 == mCurNumRects) return true; return false; }
*/

	std::list<Recta> mRects;

private:

	typedef std::list<Recta>::iterator mIterType;

/*
	int mDebugLevel;

	// STL's list is a strange beast. iterator end is ALWAYS end. even after some push_back's.
	// And so, cause the size() method can be slow, I have to keep track of the size myself.
	int mCurNumRects;
*/	
	void mCutRect(Recta& rNew, mIterType& rOld);
};


//////////////////////////////////////////////////////////////////////////////


#include "sys_include_end.h"


#endif

