// Simple timer class, by ETM.
// Allegro version 2. Uses processors time-stamp-counter. Allegro only helps
// to get current processor frequency. Well, it must be written in an allegro
// config file (must be opened), under the section named [ShouldBeGone].
//  Example config file contests.
// ....
// [ShouldBeGone]
// processor_frequency = 700
// ....
//  update: 27.01.2003


#include "Ticker.h"

#include "allegro.h"

#include "sys_include.h"

using namespace std;

// Definitely not public... :)

#ifdef _MSC_VER

    inline volatile uint64 RDTSC() {
        uint64 TSC;
        _asm rdtsc                      
        _asm lea ebx, TSC         
        _asm mov dword ptr [ebx], eax   
        _asm mov dword ptr [ebx+4], edx 
    
        return TSC;
	}

#elif defined(__GNUC__)

    inline volatile long long RDTSC(void)
	{
		long long tsc;
		__asm__ __volatile__("rdtsc" : "=A" (tsc));
		return tsc;
	}

#endif


/////////////////////////////////// PUBLIC ///////////////////////////////////

// ** START ********************* lifecycle ******************************* //


Ticker::Ticker()
{
    mElapsedTime = 0;
    mPrevMomentCount = 0;
    mFrequency = 1000 * 1000 *
        get_config_int("ShouldBeGone", "processor_frequency", 0);

    if (0 == mFrequency) {
        mFrequency = 1000 * 1000 * 800;
        loginf << "Didn't find processor frequency in current open config file. Using " 
			   << (unsigned int)(mFrequency / 1000 / 1000) << "MHz." << endl;
    }

	Tick();
}


// ** END *********************** lifecycle ******************************* //

// ** START ********************* operators ******************************* //
// ** END *********************** operators ******************************* //

// ** START ********************* operations ****************************** //


double Ticker::ToSeconds(uint64 const count) const
{
	// uint64 to double not implemented by msvc6. grrr.
    return (double)((int64)count) / (int64)(mFrequency);
}


double Ticker::GetElapsedTime(void) const
{
    return mElapsedTime;
}


double Ticker::Tick(void)
{
	uint64 cur_count = GetTimerCount(); 
	uint64 elapsed = cur_count - mPrevMomentCount;

	// !mPrevMomentCount is the same as 1111...(64 ones in binary) minus mPrevMomentCount.
	if (mPrevMomentCount > cur_count) elapsed = !mPrevMomentCount + cur_count;
    mPrevMomentCount = cur_count;

	mElapsedTime = ToSeconds(elapsed);
	return mElapsedTime;
}


// ** END *********************** operations ****************************** //

// ** START *********************** access ******************************** //


uint64 Ticker::GetTimerFrequency(void) const
{
    return mFrequency;
}


uint64 Ticker::GetTimerCount(void)
{
    mCurrentCounts = RDTSC();

    return mCurrentCounts;
}


// ** END ************************* access ******************************** //

// ** START ********************** inquiry ******************************** //
// ** END ************************ inquiry ******************************** //

///////////////////////////////// PROTECTED //////////////////////////////////

////////////////////////////////// PRIVATE ///////////////////////////////////

