// 
//  update: 22.11.03 15:41


#include "World.h"


#include <vector>
#include <stdio.h>

#include "allegro.h"

#include "iConnection.h"
#include "Canvas.h"
#include "Jelly.h"
#include "Ball.h"
#include "Cave.h"
#include "SmartPtr.h"


#include "sys_include.h"


using namespace std;


/////////////////////////////////// PUBLIC ///////////////////////////////////



vector<SmartPtr<Jelly> > jellys;
vector<SmartPtr<Cave> > caves;
Ball ball;
float camera_x;
float camera_y;
int contacts = 0; // score
int game_mode;
float power;
float power_dec;

float thrust;
float max_thrust;
float cave_thrust;
float jelly_thrust;
float freefall_thrust;

float gravity;


static float memslot_x[100];
static float memslot_y[100];


void setmem_coord(int slot, float x, float y)
{
    if (slot < 0 || slot >= 100) { logerr << "memslot out of memory" << endl; assert(false); }
    memslot_x[slot] = x;
    memslot_y[slot] = y;
}

void getmem_coord(int slot, float* x, float* y)
{
    if (slot < 0 || slot >= 100) { logerr << "memslot out of memory" << endl; assert(false); }
    *x = memslot_x[slot];
    *y = memslot_y[slot];
}


void generate_more();


// ** START ********************* lifecycle ******************************* //


World::World()
{
    ball.init();
    
    ball.x = 70;
    ball.y = 240;
    camera_x = -ball.x + SCREEN_W / 2;
    camera_y = -ball.y + SCREEN_H / 2;
    contacts = 0;
    power = 1000;
    thrust = 0;
    max_thrust = 5;

    game_mode = get_config_int("Game", "game_mode", 1);
    cave_thrust = get_config_float("Game", "cave_thrust", 1);
    jelly_thrust = get_config_float("Game", "jelly_thrust", 1);
    power_dec = get_config_float("Game", "power_dec", 1);
    gravity = get_config_float("Game", "gravity", 0.3);
    freefall_thrust = get_config_float("Game", "freefall_thrust", 0.2);

    jellys.clear();
    caves.clear();
    
    const char* filename = get_config_string("Game", "level1", "nofilefound");
    loginf << "-------------------------------------------------" << endl;
    loginf << "loading level: " << filename << endl << endl;
    
    FILE* fp = fopen(filename, "rb");
    if (!fp)
    {
        logerr << "no level file found: " << filename << endl;
        //assert(false); 
        goto jump_over_file_load;
    }

    char s[10000]; s[0] = 0; // should be enough..

    do
    {
        fscanf(fp, "%s", s);
/*    
        // read comment block..
        do {
            if (s[0] == '#') fscanf(fp, "\n"); // read whole comment-line.
            fscanf(fp, "%s", s);
        } while(s[0] == '#');
*/

        //  ----------- read different components ------------

        if (!strcmp(s, "cavesegment"))
        {
            loginf << "found: cavesegment" << endl;
            float f1, f2, f3, f4;
            fscanf(fp, "%f %f %f %f", &f1, &f2, &f3, &f4);
            
            loginf << int(f1) << " " << f2 << " " << f3 << " " << int(f4) << endl;

            SmartPtr<Cave> c(new Cave);
            caves.push_back(c);
            c->generate(int(f1), f2, f3, int(f4));

            do
            {
                fscanf(fp, "%s", s);
                if (!strcmp(s, "t"))
                {
                    fscanf(fp, "%f %f", &f1, &f2);
                    c->translate(f1, f2);
                    loginf << "translated: " << f1 << " " << f2 << endl;
                }
                else if (!strcmp(s, "r"))
                {
                    fscanf(fp, "%f", &f1);
                    c->rotate(f1);
                    loginf << "rotated: " << f1 << endl;
                }
                else if (!strcmp(s, "start>"))
                {
                    fscanf(fp, "%f", &f1);
                    float x, y;
                    c->baseStartCoord(&x, &y);
                    setmem_coord((int)f1, x, y);
                    loginf << "memory slot save: " << (int)f1 << endl;
                }
                else if (!strcmp(s, "end>"))
                {
                    fscanf(fp, "%f", &f1);
                    float x, y;
                    c->baseEndCoord(&x, &y);
                    setmem_coord((int)f1, x, y);
                    LOG(x);
                    LOG(y);
                    LOG((int)f1);
                    loginf << "memory slot save: " << (int)f1 << endl;
                }
                else if (!strcmp(s, "start<"))
                {
                    fscanf(fp, "%f", &f1);
                    float x, y;
                    c->baseStartCoord(&x, &y);
                    float x2, y2;
                    getmem_coord((int)f1, &x2, &y2);
                    c->translate(x2 - x, y2 - y);
                    loginf << "memory slot read: " << (int)f1 << endl;
                }
                else if (!strcmp(s, "end<"))
                {
                    fscanf(fp, "%f", &f1);
                    float x, y;
                    c->baseEndCoord(&x, &y);
                    float x2, y2;
                    getmem_coord((int)f1, &x2, &y2);
                    c->translate(x2 - x, y2 - y);
                    loginf << "memory slot read: " << (int)f1 << endl;
                }
                else if (strcmp(s, "!")) loginf << "unknown token: " << s << endl;
            } while (strcmp(s, "!") && !feof(fp));
            
            loginf << endl;
        }
        else if (!strcmp(s, "jelly"))
        {
            loginf << "found: jelly" << endl;
            float f1, f2, f3, f4;
            fscanf(fp, "%s %f %f %f", s, &f1, &f2, &f3);
            
            SmartPtr<Jelly> j(new Jelly);
            jellys.push_back(j);
            j->ball = &ball;

            if (!strcmp(s, "blob"))
            {
                loginf << s << " " << f1 << " " << f2 << " " << (int)f3 << endl;
                j->generateBlob(f1, f2, (int)f3);
            }
            else if (!strcmp(s, "rectangular"))
            {
                loginf << s << " " << f1 << " " << (int)f2 << " " << (int)f3 << endl;
                j->generateRectangular(f1, (int)f2, (int)f3);
            }
            else
            {
                logerr << "no known jelly type" << endl;
                assert(false);            
            }

            do
            {
                fscanf(fp, "%s", s);
                if (!strcmp(s, "t"))
                {
                    fscanf(fp, "%f %f", &f1, &f2);
                    j->translate(f1, f2);
                    loginf << "translated: " << f1 << " " << f2 << endl;
                }
                else if (!strcmp(s, "r"))
                {
                    fscanf(fp, "%f", &f1);
                    j->rotate(f1);
                    loginf << "rotated: " << f1 << endl;
                }
                else if (!strcmp(s, "start>"))
                {
                    fscanf(fp, "%f", &f1);
                    float x, y;
                    j->baseStartCoord(&x, &y);
                    setmem_coord((int)f1, x, y);
                    loginf << "memory slot save: " << (int)f1 << endl;
                }
                else if (!strcmp(s, "end>"))
                {
                    fscanf(fp, "%f", &f1);
                    float x, y;
                    j->baseEndCoord(&x, &y);
                    setmem_coord((int)f1, x, y);
                    loginf << "memory slot save: " << (int)f1 << endl;
                }
                else if (!strcmp(s, "start<"))
                {
                    fscanf(fp, "%f", &f1);
                    float x, y;
                    j->baseStartCoord(&x, &y);
                    float x2, y2;
                    getmem_coord((int)f1, &x2, &y2);
                    j->translate(x2 - x, y2 - y);
                    loginf << "memory slot read: " << (int)f1 << endl;
                }
                else if (!strcmp(s, "end<"))
                {
                    fscanf(fp, "%f", &f1);
                    float x, y;
                    j->baseEndCoord(&x, &y);
                    float x2, y2;
                    getmem_coord((int)f1, &x2, &y2);
                    j->translate(x2 - x, y2 - y);
                    loginf << "memory slot read: " << (int)f1 << endl;
                }
            } while (strcmp(s, "!") && !feof(fp));
            
            loginf << endl;
        }
        else if (!strcmp(s, "player"))
        {
            loginf << "found: player" << endl;
            float f1, f2, f3, f4;
            fscanf(fp, "%f %f", &f1, &f2);
            ball.x = f1;
            ball.y = f2;
        }
    } while (!feof(fp));
    
    
//    scanf();

    if (fp != NULL) fclose(fp);
    // fclose(NULL); // this crashes?
    loginf << "--------level loaded------------------------------------" << endl;
    
    
jump_over_file_load:

    //generate_more();
    
    if (jellys.size() == 0)
    {
        logerr << "no jellys on the level? no good." << endl;
        assert(false);
    }
}


World::~World()
{
}


// ** END *********************** lifecycle ******************************* //

// ** START ********************* operators ******************************* //
// ** END *********************** operators ******************************* //


void generate_more()
{
    // cavesegment 20 20 180 10 r 20 t 0 300 end> 0 !
    // jelly blob 200 60 10 t 600 550 r 11 start< 0 end> 0 !

    // -------------------------------
    /*{
        SmartPtr<Jelly> j(new Jelly);
        jellys.push_back(j);
    
        j->generateRectangular(20, 13, 3);
        j->translate(10, 360);
        j->ball = &ball; 
    }*/

    // create the floor
    {
    SmartPtr<Cave> c(new Cave);
    caves.push_back(c);
    c->generate(20, rand()%10 - 5 + 20, 180, 10);
    //c->rotate(rand()%6 - 3);
    c->translate(10, 390);

    float x, y;
    float x2, y2;
    c->baseEndCoord(&x, &y);

    for (int i = 0; i < 10; i++)
    {
        SmartPtr<Jelly> j(new Jelly);
        jellys.push_back(j);
    
        j->generateBlob(200, 60, 10);
//        j->rotate(rand()%6 - 3);
        j->ball = &ball;
        
        j->baseStartCoord(&x2, &y2);
        j->translate(x - x2, y - y2);      
        // read the new end-coord
        j->baseEndCoord(&x, &y);

        // cavesegment 20 20 180 10 r 20 t 0 300 end> 0 !
        SmartPtr<Cave> c(new Cave);
        caves.push_back(c);
        c->generate(20, rand()%10 - 5 + 20, 180, 10);
//        c->rotate(rand()%6 - 3);
        c->baseStartCoord(&x2, &y2);
        c->translate(x - x2, y - y2);
        // read the new end-coord
        c->baseEndCoord(&x, &y);
    }
    }
    // create the ceiling
    {
    SmartPtr<Cave> c(new Cave);
    caves.push_back(c);
    c->generate(20, rand()%10 - 5 + 20, 180, 10);
//    c->rotate(rand()%6 - 3);
    c->translate(10, 60);
    // read the new end-coord

    float x, y;
    float x2, y2;
    c->baseEndCoord(&x, &y);

    for (int i = 0; i < 10; i++)
    {
        SmartPtr<Jelly> j(new Jelly);
        jellys.push_back(j);
    
        j->generateBlob(200, 60, 10);
//        j->rotate(rand()%6 - 3 + 180);
        j->rotate(180);
        j->ball = &ball;
        
        j->baseEndCoord(&x2, &y2);
        j->translate(x - x2, y - y2);      
        // read the new end-coord
        j->baseStartCoord(&x, &y);

        // cavesegment 20 20 180 10 r 20 t 0 300 end> 0 !
        SmartPtr<Cave> c(new Cave);
        caves.push_back(c);
        c->generate(20, rand()%10 - 5 + 20, 180, 10);
//        c->rotate(rand()%6 - 3);
        c->baseStartCoord(&x2, &y2);
        c->translate(x - x2, y - y2);
        // read the new end-coord
        c->baseEndCoord(&x, &y);
    }
    }
}



void World::rightleft(float f)
{
    jellys[0]->masspoints[0].fx += f;
}


void World::updown(float f)
{
    jellys[0]->masspoints[0].fy += f;
}


void handle_keyboard()
{
    //  ********  KEYBOARD handling  *********

	//if (keypressed())
    //    clear_keybuf();
/*
	if (key[KEY_A])
        jellys[0]->masspoints[0].fx -= 10;
	if (key[KEY_D])
        jellys[0]->masspoints[0].fx += 10;
	if (key[KEY_W])
        jellys[0]->masspoints[0].fy -= 10;
	if (key[KEY_S])
        jellys[0]->masspoints[0].fy += 10;
*/
    thrust = 0;
    if (power > 0)
    {
    	if (key[KEY_LEFT])
    	{
            thrust = -max_thrust;          
            ball.fx -= ball.m * freefall_thrust / 5;
        }
    	if (key[KEY_RIGHT])
    	{
            thrust = max_thrust;
            ball.fx += ball.m * freefall_thrust / 5;
        }
    	if (key[KEY_UP])
    	{
            ball.fy -= ball.m * freefall_thrust / 5;
        }
    	if (key[KEY_DOWN])
    	{
            ball.fy += ball.m * freefall_thrust / 5;
        }
    }
    if (!game_mode)
    {
    	if (key[KEY_LEFT])
    	{
            ball.fx -= ball.m * 0.5;
        }
    	if (key[KEY_RIGHT])
    	{
            ball.fx += ball.m * 0.5;
        }
    	if (key[KEY_UP])
    	{
            ball.fy -= ball.m * 0.5;
        }
    	if (key[KEY_DOWN])
    	{
            ball.fy += ball.m * 0.5;
        }
    }
    
	if (key[KEY_SPACE])
        ball.radius += 1;
	if (key[KEY_F])
        ball.radius -= 1;
}


void accelerate_ball(float normal_x, float normal_y)
{
    ball.fx += -normal_y * thrust * ball.m;
    ball.fy += normal_x * thrust * ball.m;

    float powdec;
    if (game_mode && power > 0) powdec = power_dec; else powdec = 0;
    power -= powdec;
}


void World::update()
{
    assert(jellys.size() > 0);
    
    // camera tracking speed
    camera_x += (-camera_x - ball.x + screen->w / 2) / 20;
    camera_y += (-camera_y - ball.y + screen->h / 2) / 20;


    for (vector<SmartPtr<Cave> >::iterator i = caves.begin(); i != caves.end(); ++i)
    {
        (*i)->handleBall(&ball);
        
        if ((*i)->was_contact)
            accelerate_ball((*i)->contact_normal_x * 5 * cave_thrust, (*i)->contact_normal_y * 5 * cave_thrust);
    }

    // thats right. we do everything 5 ti,..
    for (int times = 0; times < 5; times++)
    {
        handle_keyboard();
    
        ball.fy += ball.m * gravity; // gravity

        for (vector<SmartPtr<Jelly> >::iterator i = jellys.begin(); i != jellys.end(); ++i)
        {
            (*i)->update();
            
            if ((*i)->was_contact)
                accelerate_ball((*i)->contact_normal_x * jelly_thrust, (*i)->contact_normal_y * jelly_thrust);
        }
        
        ball.update();
    }
}


void World::update(iConnection* conn)
{
    update();
}


void World::draw(Canvas* dest)
{
    assert(dest);
 
    for (vector<SmartPtr<Jelly> >::iterator i = jellys.begin(); i != jellys.end(); ++i)
    {
        (*i)->draw(dest, camera_x, camera_y);
        
        contacts += (*i)->contacts;
        power += (*i)->contacts;
        (*i)->contacts = 0;
    }
    for (vector<SmartPtr<Cave> >::iterator i = caves.begin(); i != caves.end(); ++i)
    {
        (*i)->draw(dest, camera_x, camera_y);
    }    
    ball.draw(dest, camera_x, camera_y);
    

    textprintf_ex(dest->mpGetAllegro(), font,
                  8, 8,
                  makecol(128, 90, 40),     
                  makecol(255,255,255),     
                  "score:        %i ", (contacts));
                  
    if (power < 0) power = 0;
    
    textprintf_ex(dest->mpGetAllegro(), font,
                  8, 17,
                  makecol(128, 90, 40),     
                  makecol(255,255,255),     
                  "jelly-energy: %i ", (int)power);
}


///////////////////////////////// PROTECTED //////////////////////////////////

////////////////////////////////// PRIVATE ///////////////////////////////////

