// iGraphics. Drawable object interface.
//  update: 25.09.03 21:20


#ifndef __IGRAPHICS_H__
#define __IGRAPHICS_H__


//struct BITMAP;
class Canvas;
class RectMesh;


#include "sys_include.h"

// Make an RGBA int.
#define RGBA_COL(r, g, b, a) (((r) << 24) + ((g) << 16) + ((b) << 8) + (a))

// Make an XRGB int.
#define XRGB_COL(r, g, b) (((r) << 16) + ((g) << 8) + (b))

// Extract color information from an RGBA formatted int.
#define RGBA_R(x) (((x) >> 24) & 0x000000FF)
#define RGBA_G(x) (((x) >> 16) & 0x000000FF)
#define RGBA_B(x) (((x) >> 8) & 0x000000FF)
#define RGBA_A(x) ((x) & 0x000000FF)

// Extract color information from an XRGB formatted int.
#define RGB_R(x) (((x) >> 24) & 0x000000FF)
#define RGB_G(x) (((x) >> 16) & 0x000000FF)
#define RGB_B(x) (((x) >> 8) & 0x000000FF)


// An object of class iGraphis is something you can draw.
// And sometimes also something you can draw onto (if iGraphics is also a Canvas object).

class iGraphics
{
public:

	enum E_BLEND_MODE { EUnknown, // This logs an error. Makes the picture green. And uses ECopy.
						EMulti, // Multiplicative blend
						ECopy // Pure copy. The fastest mode
	};


	iGraphics();
	virtual ~iGraphics() = 0;


	void BlendMode(E_BLEND_MODE mode);
	E_BLEND_MODE BlendMode();


	virtual float Xsize(void);
	virtual float Ysize(void);

	// Makes use of blending mode.
	// upLx, upLy are on the pixel edge. The first pixel center is on the coordinates
	// (0,5, 0.5). Draw(..) and Pixel(..) have then different coordinate systems. Off by 0.5 pixels.
	//virtual void Draw(iGraphics* pDest, float upLx, float upLy) = 0;
	//virtual void Draw(iGraphics* pDest, RectMesh& rMask, float upLx, float upLy) = 0;
	// This is the same as the next one, but with pMask = NULL.
	virtual void Draw(Canvas* pDest, float upLx, float upLy) = 0;
	// pMask = NULL <- signals that will not use the RectMesh.
	virtual void Draw(Canvas* pDest, RectMesh* pMask, float upLx, float upLy) = 0;

	// Adds to pMesh the area covered by *this. Offset x, y. To make it more
	// effective to add children, if needed.
    virtual void Coverage(RectMesh* pMesh, float x, float y);

    // Default action is to return false.
    virtual bool IsOpaque();

	//virtual void Pixel(int r, int g, int b, int a) = 0;

//protected:

	//virtual BITMAP* mpGetAllegro(void) = 0;


private:

	E_BLEND_MODE mBlendMode;
};


#include "sys_include_end.h"


#endif

