// Platform independent API for system, by ETM.
// This program is in the public domain.
//  update: 24.11.03 14:44


// NOTE: if you happen to use the allegro version of this little library,
// make sure you don't call allegro_init() in your code. or TWO ugly windows
// will appear.


// ************** EVERYTHING IN HERE IS PLATFORM-INDEPENDENT ************** //


#ifndef _SYS_FUNCTIONS_H_
#define _SYS_FUNCTIONS_H_


#include <fstream>
#include <cassert>


// ** START ******************** declarations ***************************** //


class sys_functions {
private:
public:
    sys_functions();
    ~sys_functions();
    int OSProcedure(void);
};


// ** END ********************** declarations ***************************** //

// ** START ******************** definitions ****************************** //


// Undefine these if your program is already debugged.
//#define DEBUG 1
//#define MEM_DEBUG 1 // Uses a memory-tracker. To detect leaks, you know...


#ifdef DEBUG
    #undef NDEBUG
    #undef assert // VisualC assert was not working sometimes. Well, this is better anyways.
    #if defined(_MSC_VER)
    //    #define assert(x) if (!(x)) { logerr "assert: " #x "\n"; __asm int 3; exit(1); }
		#define assert(x) if (!(x)) { logerr << "assert: " << #x << "\n"; exit(1); }
		//#define assert(x) if (!(x)) { logerr << "assert: #x \n" << std::endl; exit(1); }
		//#define assert(x) exit(1);
	#else // TODO: to insert a breakpoint (int 3) for gnu compiler too.
		#define assert(x) if (!(x)) { logerr << "assert: " << #x << std::endl; exit(1); }
	#endif
#else
    #define NDEBUG 1
#endif


#ifdef __GNUC__

    //#define __FUNCTION_NAME__ __FUNCTION__
    // __PRETTY_FUNCTION__ only supported in GNU compilers.
    #define __FUNCTION_NAME__ __PRETTY_FUNCTION__
    #define __LINE_NUM__ __LINE__
    #define __FILE_NAME__ __FILE__
    typedef long long int64;
    typedef long long uint64;

#elif defined(_other_compiler_) // insert your compiler

    // Change these for other compilers.
    #define __FUNCTION_NAME__ __FUNCTION__
    #define __LINE_NUM__ __LINE__
    #define __FILE_NAME__ __FILE__
    typedef signed __int64 int64;
    typedef unsigned __int64 uint64;

// TODO: MSVC version detection. if < 6, then output an error message.
#elif defined(_MSC_VER) // should work for MSVC compilers

    // Change these for other compilers.
    #define __FUNCTION_NAME__ "__FUNCTION__NAME__"
    #define __LINE_NUM__ __LINE__
    #define __FILE_NAME__ __FILE__
    // should be something like #ifdef MSVC...
    typedef signed __int64 int64;
    typedef unsigned __int64 uint64;

#endif


// types: ^
//     __FUNCTION_NAME__ : char*
//     __LINE_NUM__      : int
//     __FILE_NAME__     : char*


// logs errors, includes function name and line number of the error message.
// usage: logerr << "quality check failed." << endl;
//
// output: line 189 of file foo.cpp:  DoFractal(*int, *float, *float)
// output: ERROR: quality check failed.
#define logerr loginf << "\n" << "line " << __LINE_NUM__ << " of file " << \
        __FILE_NAME__ << ":  " << __FUNCTION_NAME__ << "\n" << "ERROR: "

// logs messages, includes function name and line number of the message.
// usage: logmsg << "Unusual parameter combination. I'll see what I can do..."
//               << endl;
//
// output: line 167: DoFractal(*int, *float, *float)
// output: Unusual parameter combination. I'll see what I can do...
#define logmsg loginf << "line " << __LINE_NUM__ << " of file " << \
        __FILE_NAME__ << " in function " << __FUNCTION_NAME__ << "\n"


// LOG(my_variable);
#define LOG(x) loginf << #x << " = " << (x) << endl

extern sys_functions sys;

// usage: loginf << "selected bit-depth : " << (int)bitdepth << endl;
// output: selected bit-depth : 32
// TODO: Should make the stream unbuffered, so it won't depend on << endl's
//       everywhere to flush it.

extern std::ofstream loginf;


// ** END ********************** definitions ****************************** //


#endif // _SYS_FUNCTIONS_H_

