const float PI = 3.141592654f;

#include <sstream>
#include <iomanip>

#include "allegro.h"

#include "Mouse.h"
#include "Canvas.h"
#include "Recta.h"

#include "FpsCounter.h"
#include "Magnifier.h"
#include "PulseGenerator.h"

#include "World.h"


#include "sys_include.h"


using namespace std;


// Set graphics mode. Windowed or fullscreen, bit-depth, resolution.
// Most settings will be read from an allegro configuration file.
// (no support for 8-bit mode)
void set_gr_mode(void)
{
    int col_depth = get_config_int("VideoMode", "color_depth", 32);
    int windowed = get_config_int("VideoMode", "windowed", 1);
    int x_res = get_config_int("VideoMode", "x_res", 640);
    int y_res = get_config_int("VideoMode", "y_res", 480);

    /*if (col_depth == 8) {
        logerr << "col_depth 8(256) not supported." << endl;
        exit(1);
    }*/

    set_color_depth(col_depth);

    loginf << "Trying video mode: "  << x_res << "x" << y_res << "x" <<
              col_depth << endl;

	int window_mode;
	if (windowed) window_mode = GFX_AUTODETECT_WINDOWED;
	    else window_mode = GFX_AUTODETECT;

    if (set_gfx_mode(window_mode, x_res, y_res, 0, 0) != 0) {
        logerr << allegro_error << endl;
        allegro_message("Unable to set graphic mode\n%s\n", allegro_error);
        exit(1); // Truly an abnormal exit :(
    }

    //set_palette(desktop_palette);

    loginf << "Video mode set: "  << SCREEN_W << "x" << SCREEN_H << "x" << bitmap_color_depth(screen) << endl;
}


//void m_breakOnAllocation(unsigned int count);

bool g_game_over = false;

void gameover()
{
    g_game_over = true;
    loginf << "gameover() called. ok." << endl;
}


int main(int argc, char** argv)
{
// m_breakOnAllocation(204);
{
    loginf << "Usage: 'thinkaboutmars.exe configfile.cfg'" << endl;
    loginf << "default configuration file is general.cfg" << endl << endl;

    set_window_title("..small..and..still..testing..");
    // Try to enable the little x button. Not workin :( why oh why?
    // Deprecated? no replacement? why? :( sad-sad day for me..
    //if (set_window_close_button(FALSE) !=0 )
    //{
    //    set_window_close_hook(gameover);
    //}

    install_keyboard();

    if (argc == 2) set_config_file(argv[1]); else
        set_config_file("general.cfg");


    int debug_level = get_config_int("Menu", "debug_level", 0);

    //set_alpha_blender(); ???


    set_gr_mode();
    set_display_switch_mode(SWITCH_AMNESIA); // for what???
    Mouse mouse;


    int  doublebuffer = get_config_int("Menu", "doublebuffer", 0);
    int  windowed = get_config_int("VideoMode", "windowed", 1);
    bool b_vsync = get_config_int("VideoMode", "vsync", 0) ? 1 : 0;
    float physics_frequency = get_config_float("Game", "physics_frequency", 1);


    Canvas back_buffer;
    if (doublebuffer)
    {
		back_buffer.SetAllegro( create_bitmap(SCREEN_W, SCREEN_H) );
		back_buffer.Clear(255,255,255,255);
	}

    Canvas screen_pic; // screen.
    screen_pic.SetAllegroPointer(screen);
    screen_pic.Clear(255,255,255,255);

	Recta rmagn_source;
	Recta rmagn_dest;
	{
        rmagn_dest.Pos(10, 10);
        rmagn_dest.Size(600, 140);
        rmagn_source.Pos(400, 340);
        rmagn_source.Size(300, 70);
    }
	Magnifier magnifier;

    // TODO: display frame time also. FPS shows almost no usable information.
    FpsCounter fps_counter(2);


    #define PRINT_FPS(pPic)                                                  \
        textprintf_right_ex((pPic)->mpGetAllegro(), font,                    \
                           10 + 3 * 8, SCREEN_H - 18,                        \
                           makecol(128, 90, 40),                             \
                           makecol(255,255,255),                             \
                           "  %i", (int)(fps_counter.GetFreq() + 0.5))       





    World world;
    PulseGenerator pulsegen;
    pulsegen.Freq(physics_frequency);
    SmartPtr<iConnection> worldupdateconnection;
    worldupdateconnection = pulsegen.Attach(&world);



    if (!windowed) acquire_screen(); // strange.. the if(windowed) thing...
    do
    {

            if (windowed) acquire_screen(); // strange.. the windowed thing...

        // Sometimes allegro might change it's screen address. Let's be
        // prepared. Reset our pic to the possibly new address.
        screen_pic.SetAllegroPointer(screen);

		// Fps counter stuff here.
		fps_counter.Tick();

        if (0 == doublebuffer)
        {
            mouse.Erase(&screen_pic, 0, 0);

            // CALL THE DRAW   //   scripting.Call("Update");
            //world.update();
            pulsegen.Notify();
            screen_pic.Clear(255,255,255,255);
            world.draw(&screen_pic);
            
            if (b_vsync) vsync();
            PRINT_FPS(&screen_pic);
            mouse.Poll();
            mouse.Draw(&screen_pic, 0, 0);
        }
        
        if (1 == doublebuffer)
        {
            if (b_vsync) vsync();
            // CALL THE DRAW   //   scripting.Call("Update");
            //world.update();
            pulsegen.Notify();
            back_buffer.Clear(255,255,255,255);
            world.draw(&back_buffer);
            // COPY TO BACK-BUFFER
            PRINT_FPS(&back_buffer);
            mouse.Poll();
            mouse.Draw(&back_buffer, 0, 0);
            // COPY TO SCREEN   //   
            back_buffer.Draw(&screen_pic, 0, 0);
            mouse.Erase(&back_buffer, 0, 0);
        }
        
		//mag.Magnify(back_buffer, &screen_pic, rmags, rmagd);


            if (windowed) release_screen(); // strange.. the if(windowed) thing...

        //  ********  KEYBOARD handling  *********

		if (key[KEY_ESC])
            break;


	} while (!g_game_over);
    if (!windowed) release_screen(); // strange.. the if(windowed) thing...

    worldupdateconnection->Disconnect();
}


    // This has to be called in any case. Once before exit. Skip it, and hang..
    // shit. this doesn't help... :(
    release_screen();
    clear_keybuf();
    remove_keyboard();

//    m_dumpMemoryReport();

    return 0;
}

END_OF_MAIN();

